@@ -49,8 +49,6 @@ static int method_min_max_arity(VALUE, int *max);
 
 #define IS_METHOD_PROC_IFUNC(ifunc) ((ifunc)->func == bmcall)
 
-static VALUE proc_to_s_(VALUE self, const rb_proc_t *proc);
-
 static void
 block_mark(const struct rb_block *block)
 {
@@ -266,12 +264,8 @@ rb_proc_lambda_p(VALUE procval)
 static void
 binding_free(void *ptr)
 {
-    rb_binding_t *bind;
     RUBY_FREE_ENTER("binding");
-    if (ptr) {
-	bind = ptr;
-	ruby_xfree(bind);
-    }
+    ruby_xfree(ptr);
     RUBY_FREE_LEAVE("binding");
 }
 
@@ -281,11 +275,8 @@ binding_mark(void *ptr)
     rb_binding_t *bind = ptr;
 
     RUBY_MARK_ENTER("binding");
-
     block_mark(&bind->block);
-
-    RUBY_MARK_UNLESS_NULL(bind->path);
-
+    rb_gc_mark(bind->pathobj);
     RUBY_MARK_LEAVE("binding");
 }
 
@@ -302,7 +293,7 @@ const rb_data_type_t ruby_binding_data_type = {
 	binding_free,
 	binding_memsize,
     },
-    0, 0, RUBY_TYPED_FREE_IMMEDIATELY
+    0, 0, RUBY_TYPED_WB_PROTECTED | RUBY_TYPED_FREE_IMMEDIATELY
 };
 
 VALUE
@@ -314,6 +305,7 @@ rb_binding_alloc(VALUE klass)
     return obj;
 }
 
+
 /* :nodoc: */
 static VALUE
 binding_dup(VALUE self)
@@ -322,8 +314,8 @@ binding_dup(VALUE self)
     rb_binding_t *src, *dst;
     GetBindingPtr(self, src);
     GetBindingPtr(bindval, dst);
-    dst->block = src->block;
-    dst->path = src->path;
+    rb_vm_block_copy(bindval, &dst->block, &src->block);
+    RB_OBJ_WRITE(bindval, &dst->pathobj, src->pathobj);
     dst->first_lineno = src->first_lineno;
     return bindval;
 }
@@ -340,8 +332,8 @@ binding_clone(VALUE self)
 VALUE
 rb_binding_new(void)
 {
-    rb_thread_t *th = GET_THREAD();
-    return rb_vm_make_binding(th, th->cfp);
+    rb_execution_context_t *ec = GET_EC();
+    return rb_vm_make_binding(ec, ec->cfp);
 }
 
 /*
@@ -405,6 +397,16 @@ get_local_variable_ptr(const rb_env_t **envp, ID lid)
 
 	    for (i=0; i<iseq->body->local_table_size; i++) {
 		if (iseq->body->local_table[i] == lid) {
+		    if (iseq->body->local_iseq == iseq &&
+			iseq->body->param.flags.has_block &&
+			(unsigned int)iseq->body->param.block_start == i) {
+			const VALUE *ep = env->ep;
+			if (!VM_ENV_FLAGS(ep, VM_FRAME_FLAG_MODIFIED_BLOCK_PARAM)) {
+			    RB_OBJ_WRITE(env, &env->env[i], rb_vm_bh_to_procval(GET_EC(), VM_ENV_BLOCK_HANDLER(ep)));
+			    VM_ENV_FLAGS_SET(ep, VM_FRAME_FLAG_MODIFIED_BLOCK_PARAM);
+			}
+		    }
+
 		    *envp = env;
 		    return &env->env[i];
 		}
@@ -509,7 +511,7 @@ bind_local_variable_get(VALUE bindval, VALUE sym)
     if ((ptr = get_local_variable_ptr(&env, lid)) == NULL) {
 	sym = ID2SYM(lid);
       undefined:
-	rb_name_err_raise("local variable `%1$s' not defined for %2$s",
+	rb_name_err_raise("local variable `%1$s' is not defined for %2$s",
 			  bindval, sym);
     }
 
@@ -555,7 +557,7 @@ bind_local_variable_set(VALUE bindval, VALUE sym, VALUE val)
     env = VM_ENV_ENVVAL_PTR(vm_block_ep(&bind->block));
     if ((ptr = get_local_variable_ptr(&env, lid)) == NULL) {
 	/* not found. create new env */
-	ptr = rb_binding_add_dynavars(bind, 1, &lid);
+	ptr = rb_binding_add_dynavars(bindval, bind, 1, &lid);
 	env = VM_ENV_ENVVAL_PTR(vm_block_ep(&bind->block));
     }
 
@@ -693,8 +695,8 @@ static VALUE
 proc_new(VALUE klass, int8_t is_lambda)
 {
     VALUE procval;
-    rb_thread_t *th = GET_THREAD();
-    rb_control_frame_t *cfp = th->cfp;
+    const rb_execution_context_t *ec = GET_EC();
+    rb_control_frame_t *cfp = ec->cfp;
     VALUE block_handler;
 
     if ((block_handler = rb_vm_frame_block_handler(cfp)) == VM_BLOCK_HANDLER_NONE) {
@@ -745,7 +747,7 @@ proc_new(VALUE klass, int8_t is_lambda)
 
       case block_handler_type_ifunc:
       case block_handler_type_iseq:
-	return rb_vm_make_proc_lambda(th, VM_BH_TO_CAPT_BLOCK(block_handler), klass, is_lambda);
+	return rb_vm_make_proc_lambda(ec, VM_BH_TO_CAPT_BLOCK(block_handler), klass, is_lambda);
     }
     VM_UNREACHABLE(proc_new);
     return Qnil;
@@ -882,7 +884,7 @@ rb_proc_call(VALUE self, VALUE args)
     VALUE vret;
     rb_proc_t *proc;
     GetProcPtr(self, proc);
-    vret = rb_vm_invoke_proc(GET_THREAD(), proc,
+    vret = rb_vm_invoke_proc(GET_EC(), proc,
 			     check_argc(RARRAY_LEN(args)), RARRAY_CONST_PTR(args),
 			     VM_BLOCK_HANDLER_NONE);
     RB_GC_GUARD(self);
@@ -899,11 +901,11 @@ proc_to_block_handler(VALUE procval)
 VALUE
 rb_proc_call_with_block(VALUE self, int argc, const VALUE *argv, VALUE passed_procval)
 {
-    rb_thread_t *th = GET_THREAD();
+    rb_execution_context_t *ec = GET_EC();
     VALUE vret;
     rb_proc_t *proc;
     GetProcPtr(self, proc);
-    vret = rb_vm_invoke_proc(th, proc, argc, argv, proc_to_block_handler(passed_procval));
+    vret = rb_vm_invoke_proc(ec, proc, argc, argv, proc_to_block_handler(passed_procval));
     RB_GC_GUARD(self);
     return vret;
 }
@@ -920,7 +922,7 @@ rb_proc_call_with_block(VALUE self, int argc, const VALUE *argv, VALUE passed_pr
  *  number of mandatory arguments, with the exception for blocks that
  *  are not lambdas and have only a finite number of optional arguments;
  *  in this latter case, returns n.
- *  Keywords arguments will considered as a single additional argument,
+ *  Keyword arguments will be considered as a single additional argument,
  *  that argument being mandatory if any keyword argument is mandatory.
  *  A <code>proc</code> with no argument declarations
  *  is the same as a block declaring <code>||</code> as its arguments.
@@ -936,16 +938,16 @@ rb_proc_call_with_block(VALUE self, int argc, const VALUE *argv, VALUE passed_pr
  *     proc { |x:, y:, z:0| }.arity   #=>  1
  *     proc { |*a, x:, y:0| }.arity   #=> -2
  *
- *     proc   { |x=0| }.arity         #=>  0
- *     lambda { |x=0| }.arity         #=> -1
- *     proc   { |x=0, y| }.arity      #=>  1
- *     lambda { |x=0, y| }.arity      #=> -2
- *     proc   { |x=0, y=0| }.arity    #=>  0
- *     lambda { |x=0, y=0| }.arity    #=> -1
- *     proc   { |x, y=0| }.arity      #=>  1
- *     lambda { |x, y=0| }.arity      #=> -2
- *     proc   { |(x, y), z=0| }.arity #=>  1
- *     lambda { |(x, y), z=0| }.arity #=> -2
+ *     proc   { |a=0| }.arity         #=>  0
+ *     lambda { |a=0| }.arity         #=> -1
+ *     proc   { |a=0, b| }.arity      #=>  1
+ *     lambda { |a=0, b| }.arity      #=> -2
+ *     proc   { |a=0, b=0| }.arity    #=>  0
+ *     lambda { |a=0, b=0| }.arity    #=> -1
+ *     proc   { |a, b=0| }.arity      #=>  1
+ *     lambda { |a, b=0| }.arity      #=> -2
+ *     proc   { |(a, b), c=0| }.arity #=>  1
+ *     lambda { |(a, b), c=0| }.arity #=> -2
  *     proc   { |a, x:0, y:0| }.arity #=>  1
  *     lambda { |a, x:0, y:0| }.arity #=> -2
  */
@@ -970,11 +972,13 @@ rb_iseq_min_max_arity(const rb_iseq_t *iseq, int *max)
 static int
 rb_vm_block_min_max_arity(const struct rb_block *block, int *max)
 {
+  again:
     switch (vm_block_type(block)) {
       case block_type_iseq:
-	return rb_iseq_min_max_arity(block->as.captured.code.iseq, max);
+	return rb_iseq_min_max_arity(rb_iseq_check(block->as.captured.code.iseq), max);
       case block_type_proc:
-	return rb_vm_block_min_max_arity(vm_proc_block(block->as.proc), max);
+	block = vm_proc_block(block->as.proc);
+	goto again;
       case block_type_ifunc:
 	{
 	    const struct vm_ifunc *ifunc = block->as.captured.code.ifunc;
@@ -1042,8 +1046,8 @@ int
 rb_block_arity(void)
 {
     int min, max;
-    rb_thread_t *th = GET_THREAD();
-    rb_control_frame_t *cfp = th->cfp;
+    const rb_execution_context_t *ec = GET_EC();
+    rb_control_frame_t *cfp = ec->cfp;
     VALUE block_handler = rb_vm_frame_block_handler(cfp);
     struct rb_block block;
 
@@ -1075,8 +1079,8 @@ rb_block_arity(void)
 int
 rb_block_min_max_arity(int *max)
 {
-    rb_thread_t *th = GET_THREAD();
-    rb_control_frame_t *cfp = th->cfp;
+    const rb_execution_context_t *ec = GET_EC();
+    rb_control_frame_t *cfp = ec->cfp;
     VALUE block_handler = rb_vm_frame_block_handler(cfp);
     struct rb_block block;
 
@@ -1130,7 +1134,7 @@ iseq_location(const rb_iseq_t *iseq)
 
     if (!iseq) return Qnil;
     rb_iseq_check(iseq);
-    loc[0] = iseq->body->location.path;
+    loc[0] = rb_iseq_path(iseq);
     loc[1] = iseq->body->location.first_lineno;
 
     return rb_ary_new4(2, loc);
@@ -1250,6 +1254,39 @@ proc_hash(VALUE self)
     return ST2FIX(hash);
 }
 
+VALUE
+rb_block_to_s(VALUE self, const struct rb_block *block, const char *additional_info)
+{
+    VALUE cname = rb_obj_class(self);
+    VALUE str = rb_sprintf("#<%"PRIsVALUE":", cname);
+
+  again:
+    switch (vm_block_type(block)) {
+      case block_type_proc:
+	block = vm_proc_block(block->as.proc);
+	goto again;
+      case block_type_iseq:
+	{
+	    const rb_iseq_t *iseq = rb_iseq_check(block->as.captured.code.iseq);
+	    rb_str_catf(str, "%p@%"PRIsVALUE":%d", (void *)self,
+			rb_iseq_path(iseq),
+			FIX2INT(iseq->body->location.first_lineno));
+	}
+	break;
+      case block_type_symbol:
+	rb_str_catf(str, "%p(&%+"PRIsVALUE")", (void *)self, block->as.symbol);
+	break;
+      case block_type_ifunc:
+	rb_str_catf(str, "%p", block->as.captured.code.ifunc);
+	break;
+    }
+
+    if (additional_info) rb_str_cat_cstr(str, additional_info);
+    rb_str_cat_cstr(str, ">");
+    OBJ_INFECT_RAW(str, self);
+    return str;
+}
+
 /*
  * call-seq:
  *   prc.to_s   -> string
@@ -1263,47 +1300,7 @@ proc_to_s(VALUE self)
 {
     const rb_proc_t *proc;
     GetProcPtr(self, proc);
-    return proc_to_s_(self, proc);
-}
-
-static VALUE
-proc_to_s_(VALUE self, const rb_proc_t *proc)
-{
-    VALUE str = 0;
-    const char *cname = rb_obj_classname(self);
-    const struct rb_block *block;
-    const char *is_lambda;
-
-    block = &proc->block;
-    is_lambda = proc->is_lambda ? " (lambda)" : "";
-
-  again:
-    switch (vm_block_type(block)) {
-      case block_type_proc:
-	block = vm_proc_block(block->as.proc);
-	goto again;
-      case block_type_iseq:
-	{
-	    const rb_iseq_t *iseq = rb_iseq_check(block->as.captured.code.iseq);
-	    str = rb_sprintf("#<%s:%p@%"PRIsVALUE":%d%s>", cname, (void *)self,
-			     iseq->body->location.path,
-			     FIX2INT(iseq->body->location.first_lineno), is_lambda);
-	}
-	break;
-      case block_type_symbol:
-	str = rb_sprintf("#<%s:%p(&%+"PRIsVALUE")%s>", cname, (void *)self,
-			 block->as.symbol, is_lambda);
-	break;
-      case block_type_ifunc:
-	str = rb_sprintf("#<%s:%p%s>", cname, proc->block.as.captured.code.ifunc,
-			 is_lambda);
-	break;
-    }
-
-    if (OBJ_TAINTED(self)) {
-	OBJ_TAINT(str);
-    }
-    return str;
+    return rb_block_to_s(self, &proc->block, proc->is_lambda ? " (lambda)" : NULL);
 }
 
 /*
@@ -1473,7 +1470,7 @@ method_entry_defined_class(const rb_method_entry_t *me)
 
 /**********************************************************************
  *
- * Document-class : Method
+ * Document-class: Method
  *
  *  Method objects are created by <code>Object#method</code>, and are
  *  associated with a particular object (not just with a class). They
@@ -1619,6 +1616,12 @@ method_name(VALUE obj)
  *     meth.original_name    -> symbol
  *
  *  Returns the original name of the method.
+ *
+ *    class C
+ *      def foo; end
+ *      alias bar foo
+ *    end
+ *    C.instance_method(:bar).original_name # => :foo
  */
 
 static VALUE
@@ -1766,21 +1769,23 @@ VALUE
 rb_obj_singleton_method(VALUE obj, VALUE vid)
 {
     const rb_method_entry_t *me;
-    VALUE klass;
+    VALUE klass = rb_singleton_class_get(obj);
     ID id = rb_check_id(&vid);
 
-    if (!id) {
-	if (!NIL_P(klass = rb_singleton_class_get(obj)) &&
-	    respond_to_missing_p(klass, obj, vid, FALSE)) {
-	    id = rb_intern_str(vid);
-	    return mnew_missing(klass, obj, id, rb_cMethod);
-	}
+    if (NIL_P(klass) || NIL_P(klass = RCLASS_ORIGIN(klass))) {
       undef:
 	rb_name_err_raise("undefined singleton method `%1$s' for `%2$s'",
 			  obj, vid);
     }
-    if (NIL_P(klass = rb_singleton_class_get(obj)) ||
-	UNDEFINED_METHOD_ENTRY_P(me = rb_method_entry_at(klass, id)) ||
+    if (!id) {
+	if (respond_to_missing_p(klass, obj, vid, FALSE)) {
+	    id = rb_intern_str(vid);
+	    return mnew_missing(klass, obj, id, rb_cMethod);
+	}
+	goto undef;
+    }
+    me = rb_method_entry_at(klass, id);
+    if (UNDEFINED_METHOD_ENTRY_P(me) ||
 	UNDEFINED_REFINED_METHOD_P(me->def)) {
 	vid = ID2SYM(id);
 	goto undef;
@@ -1905,8 +1910,8 @@ rb_mod_define_method(int argc, VALUE *argv, VALUE mod)
 #if PROC_NEW_REQUIRES_BLOCK
 	body = rb_block_lambda();
 #else
-	rb_thread_t *th = GET_THREAD();
-	VALUE block_handler = rb_vm_frame_block_handler(th->cfp);
+	const rb_execution_context_t *ec = GET_EC();
+	VALUE block_handler = rb_vm_frame_block_handler(ec->cfp);
 	if (block_handler == VM_BLOCK_HANDLER_NONE) rb_raise(rb_eArgError, proc_without_block);
 
 	switch (vm_block_handler_type(block_handler)) {
@@ -1918,7 +1923,7 @@ rb_mod_define_method(int argc, VALUE *argv, VALUE mod)
 	    break;
 	  case block_handler_type_iseq:
 	  case block_handler_type_ifunc:
-	    body = rb_vm_make_proc_lambda(th, VM_BH_TO_CAPT_BLOCK(block_handler), rb_cProc, TRUE);
+	    body = rb_vm_make_lambda(ec, VM_BH_TO_CAPT_BLOCK(block_handler), rb_cProc);
 	}
 #endif
     }
@@ -1950,7 +1955,7 @@ rb_mod_define_method(int argc, VALUE *argv, VALUE mod)
 	    else {
 		rb_raise(rb_eTypeError,
 			 "bind argument must be a subclass of % "PRIsVALUE,
-			 rb_class_name(method->me->owner));
+			 method->me->owner);
 	    }
 	}
 	rb_method_entry_set(mod, id, method->me, scope_visi->method_visi);
@@ -2093,32 +2098,32 @@ method_callable_method_entry(const struct METHOD *data)
 }
 
 static inline VALUE
-call_method_data(rb_thread_t *th, const struct METHOD *data,
+call_method_data(rb_execution_context_t *ec, const struct METHOD *data,
 		 int argc, const VALUE *argv, VALUE passed_procval)
 {
-    vm_passed_block_handler_set(th, proc_to_block_handler(passed_procval));
-    return rb_vm_call(th, data->recv, data->me->called_id, argc, argv,
+    vm_passed_block_handler_set(ec, proc_to_block_handler(passed_procval));
+    return rb_vm_call(ec, data->recv, data->me->called_id, argc, argv,
 		      method_callable_method_entry(data));
 }
 
 static VALUE
-call_method_data_safe(rb_thread_t *th, const struct METHOD *data,
+call_method_data_safe(rb_execution_context_t *ec, const struct METHOD *data,
 		      int argc, const VALUE *argv, VALUE passed_procval,
 		      int safe)
 {
     VALUE result = Qnil;	/* OK */
-    int state;
+    enum ruby_tag_type state;
 
-    TH_PUSH_TAG(th);
-    if ((state = TH_EXEC_TAG()) == 0) {
+    EC_PUSH_TAG(ec);
+    if ((state = EC_EXEC_TAG()) == TAG_NONE) {
 	/* result is used only if state == 0, no exceptions is caught. */
 	/* otherwise it doesn't matter even if clobbered. */
-	NO_CLOBBERED(result) = call_method_data(th, data, argc, argv, passed_procval);
+	NO_CLOBBERED(result) = call_method_data(ec, data, argc, argv, passed_procval);
     }
-    TH_POP_TAG();
+    EC_POP_TAG();
     rb_set_safe_level_force(safe);
     if (state)
-	TH_JUMP_TAG(th, state);
+	EC_JUMP_TAG(ec, state);
     return result;
 }
 
@@ -2126,7 +2131,7 @@ VALUE
 rb_method_call_with_block(int argc, const VALUE *argv, VALUE method, VALUE passed_procval)
 {
     const struct METHOD *data;
-    rb_thread_t *const th = GET_THREAD();
+    rb_execution_context_t *ec = GET_EC();
 
     TypedData_Get_Struct(method, struct METHOD, &method_data_type, data);
     if (data->recv == Qundef) {
@@ -2137,10 +2142,10 @@ rb_method_call_with_block(int argc, const VALUE *argv, VALUE method, VALUE passe
 	int safe = rb_safe_level();
 	if (safe < safe_level_to_run) {
 	    rb_set_safe_level_force(safe_level_to_run);
-	    return call_method_data_safe(th, data, argc, argv, passed_procval, safe);
+	    return call_method_data_safe(ec, data, argc, argv, passed_procval, safe);
 	}
     }
-    return call_method_data(th, data, argc, argv, passed_procval);
+    return call_method_data(ec, data, argc, argv, passed_procval);
 }
 
 /**********************************************************************
@@ -2252,7 +2257,7 @@ umethod_bind(VALUE method, VALUE recv)
 	}
 	else {
 	    rb_raise(rb_eTypeError, "bind argument must be an instance of % "PRIsVALUE,
-		     rb_class_name(methclass));
+		     methclass);
 	}
     }
 
@@ -2287,6 +2292,7 @@ rb_method_entry_min_max_arity(const rb_method_entry_t *me, int *max)
 {
     const rb_method_definition_t *def = me->def;
 
+  again:
     if (!def) return *max = 0;
     switch (def->type) {
       case VM_METHOD_TYPE_CFUNC:
@@ -2303,13 +2309,12 @@ rb_method_entry_min_max_arity(const rb_method_entry_t *me, int *max)
       case VM_METHOD_TYPE_IVAR:
 	return *max = 0;
       case VM_METHOD_TYPE_ALIAS:
-	return rb_method_entry_min_max_arity(def->body.alias.original_me, max);
+	def = def->body.alias.original_me->def;
+	goto again;
       case VM_METHOD_TYPE_BMETHOD:
 	return rb_proc_min_max_arity(def->body.proc, max);
-      case VM_METHOD_TYPE_ISEQ: {
-	const rb_iseq_t *iseq = rb_iseq_check(def->body.iseq.iseqptr);
-	return rb_iseq_min_max_arity(iseq, max);
-      }
+      case VM_METHOD_TYPE_ISEQ:
+	return rb_iseq_min_max_arity(rb_iseq_check(def->body.iseq.iseqptr), max);
       case VM_METHOD_TYPE_UNDEF:
       case VM_METHOD_TYPE_NOTIMPLEMENTED:
 	return *max = 0;
@@ -2351,8 +2356,10 @@ rb_method_entry_arity(const rb_method_entry_t *me)
  *  Returns an indication of the number of arguments accepted by a
  *  method. Returns a nonnegative integer for methods that take a fixed
  *  number of arguments. For Ruby methods that take a variable number of
- *  arguments, returns -n-1, where n is the number of required
- *  arguments. For methods written in C, returns -1 if the call takes a
+ *  arguments, returns -n-1, where n is the number of required arguments.
+ *  Keyword arguments will be considered as a single additional argument,
+ *  that argument being mandatory if any keyword argument is mandatory.
+ *  For methods written in C, returns -1 if the call takes a
  *  variable number of arguments.
  *
  *     class C
@@ -2362,6 +2369,10 @@ rb_method_entry_arity(const rb_method_entry_t *me)
  *       def four(a, b); end
  *       def five(a, b, *c);    end
  *       def six(a, b, *c, &d); end
+ *       def seven(a, b, x:0); end
+ *       def eight(x:, y:); end
+ *       def nine(x:, y:, **z); end
+ *       def ten(*a, x:, y:); end
  *     end
  *     c = C.new
  *     c.method(:one).arity     #=> 0
@@ -2370,6 +2381,10 @@ rb_method_entry_arity(const rb_method_entry_t *me)
  *     c.method(:four).arity    #=> 2
  *     c.method(:five).arity    #=> -3
  *     c.method(:six).arity     #=> -3
+ *     c.method(:seven).arity   #=> -3
+ *     c.method(:eight).arity   #=> 1
+ *     c.method(:nine).arity    #=> 1
+ *     c.method(:ten).arity     #=> -2
  *
  *     "cat".method(:size).arity      #=> 0
  *     "cat".method(:replace).arity   #=> 1
@@ -2575,16 +2590,13 @@ method_inspect(VALUE method)
 {
     struct METHOD *data;
     VALUE str;
-    const char *s;
     const char *sharp = "#";
     VALUE mklass;
     VALUE defined_class;
 
     TypedData_Get_Struct(method, struct METHOD, &method_data_type, data);
-    str = rb_str_buf_new2("#<");
-    s = rb_obj_classname(method);
-    rb_str_buf_cat2(str, s);
-    rb_str_buf_cat2(str, ": ");
+    str = rb_sprintf("#<% "PRIsVALUE": ", rb_obj_class(method));
+    OBJ_INFECT_RAW(str, method);
 
     mklass = data->klass;
 
@@ -2618,11 +2630,9 @@ method_inspect(VALUE method)
 	}
     }
     else {
-	rb_str_buf_append(str, rb_class_name(mklass));
+	rb_str_buf_append(str, rb_inspect(mklass));
 	if (defined_class != mklass) {
-	    rb_str_buf_cat2(str, "(");
-	    rb_str_buf_append(str, rb_class_name(defined_class));
-	    rb_str_buf_cat2(str, ")");
+	    rb_str_catf(str, "(% "PRIsVALUE")", defined_class);
 	}
     }
     rb_str_buf_cat2(str, sharp);
@@ -2836,19 +2846,20 @@ proc_binding(VALUE self)
 
     bindval = rb_binding_alloc(rb_cBinding);
     GetBindingPtr(bindval, bind);
-
-    bind->block.as.captured.self = binding_self;
-    bind->block.as.captured.code.iseq = env->iseq;
-    bind->block.as.captured.ep = env->ep;
+    RB_OBJ_WRITE(bindval, &bind->block.as.captured.self, binding_self);
+    RB_OBJ_WRITE(bindval, &bind->block.as.captured.code.iseq, env->iseq);
+    rb_vm_block_ep_update(bindval, &bind->block, env->ep);
+    RB_OBJ_WRITTEN(bindval, Qundef, VM_ENV_ENVVAL(env->ep));
 
     if (iseq) {
 	rb_iseq_check(iseq);
-	bind->path = iseq->body->location.path;
+	RB_OBJ_WRITE(bindval, &bind->pathobj, iseq->body->location.pathobj);
 	bind->first_lineno = FIX2INT(rb_iseq_first_lineno(iseq));
     }
     else {
-	bind->path = Qnil;
-	bind->first_lineno = 0;
+	RB_OBJ_WRITE(bindval, &bind->pathobj,
+		     rb_iseq_pathobj_new(rb_fstring_cstr("(binding)"), Qnil));
+	bind->first_lineno = 1;
     }
 
     return bindval;
@@ -3116,6 +3127,7 @@ Init_Proc(void)
     rb_define_method(rb_cMethod, "hash", method_hash, 0);
     rb_define_method(rb_cMethod, "clone", method_clone, 0);
     rb_define_method(rb_cMethod, "call", rb_method_call, -1);
+    rb_define_method(rb_cMethod, "===", rb_method_call, -1);
     rb_define_method(rb_cMethod, "curry", rb_method_curry, -1);
     rb_define_method(rb_cMethod, "[]", rb_method_call, -1);
     rb_define_method(rb_cMethod, "arity", method_arity_m, 0);
@@ -3156,7 +3168,7 @@ Init_Proc(void)
     /* Module#*_method */
     rb_define_method(rb_cModule, "instance_method", rb_mod_instance_method, 1);
     rb_define_method(rb_cModule, "public_instance_method", rb_mod_public_instance_method, 1);
-    rb_define_private_method(rb_cModule, "define_method", rb_mod_define_method, -1);
+    rb_define_method(rb_cModule, "define_method", rb_mod_define_method, -1);
 
     /* Kernel */
     rb_define_method(rb_mKernel, "define_singleton_method", rb_obj_define_method, -1);
